/**
 * @jest-environment jsdom
 */

import * as governanceApi from "$lib/api/sns-governance.api";
import * as api from "$lib/api/sns.api";
import { E8S_PER_ICP } from "$lib/constants/icp.constants";
import { HOTKEY_PERMISSIONS } from "$lib/constants/sns-neurons.constants";
import { snsTokenSymbolSelectedStore } from "$lib/derived/sns/sns-token-symbol-selected.store";
import { loadSnsAccounts } from "$lib/services/sns-accounts.services";
import * as services from "$lib/services/sns-neurons.services";
import {
  disburse,
  increaseStakeNeuron,
  stakeMaturity,
  startDissolving,
  stopDissolving,
  toggleAutoStakeMaturity,
  updateDelay,
} from "$lib/services/sns-neurons.services";
import { snsNeuronsStore } from "$lib/stores/sns-neurons.store";
import { toastsError } from "$lib/stores/toasts.store";
import { transactionsFeesStore } from "$lib/stores/transaction-fees.store";
import {
  getSnsNeuronIdAsHexString,
  subaccountToHexString,
} from "$lib/utils/sns-neuron.utils";
import { numberToE8s } from "$lib/utils/token.utils";
import { bytesToHexString } from "$lib/utils/utils";
import { decodeIcrcAccount } from "@dfinity/ledger";
import { Principal } from "@dfinity/principal";
import {
  neuronSubaccount,
  type SnsNeuron,
  type SnsNeuronId,
} from "@dfinity/sns";
import {
  arrayOfNumberToUint8Array,
  fromDefinedNullable,
  fromNullable,
} from "@dfinity/utils";
import { waitFor } from "@testing-library/svelte";
import { tick } from "svelte";
import { get } from "svelte/store";
import { mockIdentity, mockPrincipal } from "../../mocks/auth.store.mock";
import { mockSnsMainAccount } from "../../mocks/sns-accounts.mock";
import {
  buildMockSnsNeuronsStoreSubscribe,
  mockSnsNeuron,
} from "../../mocks/sns-neurons.mock";
import { mockTokenStore } from "../../mocks/sns-projects.mock";

const {
  syncSnsNeurons,
  getSnsNeuron,
  addHotkey,
  removeHotkey,
  splitNeuron,
  stakeNeuron,
  loadNeurons,
  addFollowee,
} = services;

jest.mock("$lib/stores/toasts.store", () => {
  return {
    toastsError: jest.fn(),
  };
});

jest.mock("$lib/services/sns-accounts.services", () => {
  return {
    loadSnsAccounts: jest.fn(),
  };
});

describe("sns-neurons-services", () => {
  describe("syncSnsNeurons", () => {
    beforeEach(() => {
      jest.clearAllMocks();
      snsNeuronsStore.reset();
      jest.spyOn(console, "error").mockImplementation(() => undefined);
    });

    it("should call api.querySnsNeurons and load neurons in store", async () => {
      const subaccount: Uint8Array = neuronSubaccount({
        controller: mockIdentity.getPrincipal(),
        index: 0,
      });
      const neuronId: SnsNeuronId = { id: subaccount };
      const neuron = {
        ...mockSnsNeuron,
        id: [neuronId] as [SnsNeuronId],
      };
      const spyQuery = jest
        .spyOn(api, "querySnsNeurons")
        .mockImplementation(() => Promise.resolve([neuron]));
      const spyNeuronBalance = jest
        .spyOn(governanceApi, "getNeuronBalance")
        .mockImplementationOnce(() =>
          Promise.resolve(mockSnsNeuron.cached_neuron_stake_e8s)
        )
        .mockImplementation(() => Promise.resolve(BigInt(0)));
      await syncSnsNeurons(mockPrincipal);

      await tick();
      const store = get(snsNeuronsStore);
      expect(store[mockPrincipal.toText()]?.neurons).toHaveLength(1);
      expect(spyQuery).toBeCalled();
      expect(spyNeuronBalance).toBeCalled();
    });

    it("should refresh and refetch the neuron if balance doesn't match", async () => {
      const subaccount: Uint8Array = neuronSubaccount({
        controller: mockIdentity.getPrincipal(),
        index: 0,
      });
      const neuronId: SnsNeuronId = { id: subaccount };
      const neuron = {
        ...mockSnsNeuron,
        id: [neuronId] as [SnsNeuronId],
      };
      const spyQuery = jest
        .spyOn(api, "querySnsNeurons")
        .mockImplementation(() => Promise.resolve([neuron]));
      const spyNeuronQuery = jest
        .spyOn(api, "getSnsNeuron")
        .mockImplementation(() => Promise.resolve(neuron));
      const spyNeuronBalance = jest
        .spyOn(governanceApi, "getNeuronBalance")
        .mockImplementationOnce(() =>
          Promise.resolve(
            mockSnsNeuron.cached_neuron_stake_e8s + BigInt(10_000)
          )
        )
        .mockImplementation(() => Promise.resolve(BigInt(0)));
      const spyRefreshNeuron = jest
        .spyOn(governanceApi, "refreshNeuron")
        .mockImplementation(() => Promise.resolve(undefined));
      await syncSnsNeurons(mockPrincipal);

      await tick();
      const store = get(snsNeuronsStore);
      expect(store[mockPrincipal.toText()]?.neurons).toHaveLength(1);
      expect(spyQuery).toBeCalled();
      expect(spyNeuronBalance).toBeCalled();
      expect(spyRefreshNeuron).toBeCalled();
      expect(spyNeuronQuery).toBeCalled();
    });

    it("should claim neuron if find a subaccount without neuron", async () => {
      const subaccount: Uint8Array = neuronSubaccount({
        controller: mockIdentity.getPrincipal(),
        index: 1,
      });
      const neuronId: SnsNeuronId = { id: subaccount };
      const neuron = {
        ...mockSnsNeuron,
        id: [neuronId] as [SnsNeuronId],
      };
      const spyQuery = jest
        .spyOn(api, "querySnsNeurons")
        .mockImplementation(() => Promise.resolve([neuron]));
      const spyNeuronQuery = jest
        .spyOn(api, "getSnsNeuron")
        .mockImplementation(() => Promise.resolve(neuron));
      const spyNeuronBalance = jest
        .spyOn(governanceApi, "getNeuronBalance")
        .mockImplementationOnce(() =>
          Promise.resolve(mockSnsNeuron.cached_neuron_stake_e8s)
        )
        .mockImplementationOnce(() => Promise.resolve(BigInt(200_000_000)))
        .mockImplementation(() => Promise.resolve(BigInt(0)));
      const spyClaimNeuron = jest
        .spyOn(governanceApi, "claimNeuron")
        .mockImplementation(() => Promise.resolve(undefined));
      await syncSnsNeurons(mockPrincipal);

      await tick();
      await tick();
      const store = get(snsNeuronsStore);
      expect(store[mockPrincipal.toText()]?.neurons).toHaveLength(1);
      expect(spyQuery).toBeCalled();
      expect(spyNeuronBalance).toBeCalled();
      expect(spyClaimNeuron).toBeCalled();
      expect(spyNeuronQuery).toBeCalled();
    });

    it("should empty store if update call fails", async () => {
      snsNeuronsStore.setNeurons({
        rootCanisterId: mockPrincipal,
        neurons: [mockSnsNeuron],
        certified: true,
      });
      const spyQuery = jest
        .spyOn(api, "querySnsNeurons")
        .mockImplementation(() => Promise.reject(undefined));

      await syncSnsNeurons(mockPrincipal);

      await tick();
      const store = get(snsNeuronsStore);
      expect(store[mockPrincipal.toText()]).toBeUndefined();
      expect(spyQuery).toBeCalled();
    });
  });

  describe("loadNeurons", () => {
    beforeEach(() => {
      jest.clearAllMocks();
      snsNeuronsStore.reset();
      jest.spyOn(console, "error").mockImplementation(() => undefined);
    });

    it("should call api.querySnsNeurons and load neurons in store", async () => {
      const subaccount: Uint8Array = neuronSubaccount({
        controller: mockIdentity.getPrincipal(),
        index: 0,
      });
      const neuronId: SnsNeuronId = { id: subaccount };
      const neuron = {
        ...mockSnsNeuron,
        id: [neuronId] as [SnsNeuronId],
      };
      const spyQuery = jest
        .spyOn(api, "querySnsNeurons")
        .mockImplementation(() => Promise.resolve([neuron]));
      await loadNeurons({ rootCanisterId: mockPrincipal, certified: true });

      await tick();
      const store = get(snsNeuronsStore);
      expect(store[mockPrincipal.toText()]?.neurons).toHaveLength(1);
      expect(spyQuery).toBeCalled();
    });
  });

  describe("getSnsNeuron", () => {
    beforeEach(() => {
      jest.clearAllMocks();
      snsNeuronsStore.reset();
      jest.spyOn(console, "error").mockImplementation(() => undefined);
    });
    it("should call api.querySnsNeuron and call load neuron when neuron not in store", (done) => {
      const subaccount: Uint8Array = neuronSubaccount({
        controller: mockIdentity.getPrincipal(),
        index: 0,
      });
      const neuronId: SnsNeuronId = { id: subaccount };
      const neuron = {
        ...mockSnsNeuron,
        id: [neuronId] as [SnsNeuronId],
      };
      const spyNeuronBalance = jest
        .spyOn(governanceApi, "getNeuronBalance")
        .mockImplementationOnce(() =>
          Promise.resolve(mockSnsNeuron.cached_neuron_stake_e8s)
        )
        .mockImplementation(() => Promise.resolve(BigInt(0)));
      const spyQuery = jest
        .spyOn(api, "getSnsNeuron")
        .mockImplementation(() => Promise.resolve(neuron));
      const onLoad = async ({
        neuron: neuronToLoad,
        certified,
      }: {
        neuron: SnsNeuron;
        certified: boolean;
      }) => {
        expect(spyQuery).toBeCalled();
        expect(neuronToLoad).toEqual(neuron);
        if (certified) {
          await waitFor(() => expect(spyNeuronBalance).toBeCalled());
          done();
        }
      };
      getSnsNeuron({
        neuronIdHex: bytesToHexString(
          Array.from(mockSnsNeuron.id[0]?.id as Uint8Array)
        ),
        rootCanisterId: mockPrincipal,
        onLoad,
      });
    });

    it("should refresh neuron if balance does not match and load again", (done) => {
      const subaccount: Uint8Array = neuronSubaccount({
        controller: mockIdentity.getPrincipal(),
        index: 0,
      });
      const neuronId: SnsNeuronId = { id: subaccount };
      const neuron = {
        ...mockSnsNeuron,
        id: [neuronId] as [SnsNeuronId],
      };
      const stake = neuron.cached_neuron_stake_e8s + BigInt(10_000);
      const updatedNeuron = {
        ...neuron,
        cached_neuron_stake_e8s: stake,
      };
      const spyNeuronBalance = jest
        .spyOn(governanceApi, "getNeuronBalance")
        .mockImplementationOnce(() => Promise.resolve(stake))
        .mockImplementation(() => Promise.resolve(BigInt(0)));
      const spyQuery = jest
        .spyOn(api, "getSnsNeuron")
        // First is the query call and returns old neuron
        .mockImplementationOnce(() => Promise.resolve(neuron))
        // Second is the update call and returns old neuron. It will be checked.
        .mockImplementationOnce(() => Promise.resolve(neuron))
        // After refreshing we get the updated neuron.
        .mockImplementation(() => Promise.resolve(updatedNeuron));
      const spyRefreshNeuron = jest
        .spyOn(governanceApi, "refreshNeuron")
        .mockImplementation(() => Promise.resolve(undefined));
      const onLoad = ({
        neuron: neuronToLoad,
      }: {
        neuron: SnsNeuron;
        certified: boolean;
      }) => {
        // Wait until we get the updated neuron to finish the test.
        if (neuronToLoad.cached_neuron_stake_e8s === stake) {
          expect(spyQuery).toBeCalledTimes(3);
          expect(spyNeuronBalance).toBeCalledTimes(1);
          expect(spyRefreshNeuron).toBeCalledTimes(1);
          expect(neuronToLoad).toEqual(updatedNeuron);
          done();
        }
      };
      getSnsNeuron({
        neuronIdHex: bytesToHexString(
          Array.from(mockSnsNeuron.id[0]?.id as Uint8Array)
        ),
        rootCanisterId: mockPrincipal,
        onLoad,
      });
    });

    it("should return neuron if it's in the store", (done) => {
      const spyQuery = jest
        .spyOn(api, "getSnsNeuron")
        .mockImplementation(() => Promise.resolve(mockSnsNeuron));
      snsNeuronsStore.setNeurons({
        rootCanisterId: mockPrincipal,
        neurons: [mockSnsNeuron],
        certified: true,
      });
      const onLoad = ({
        neuron,
        certified,
      }: {
        neuron: SnsNeuron;
        certified: boolean;
      }) => {
        expect(spyQuery).not.toBeCalled();
        expect(neuron).toEqual(mockSnsNeuron);
        if (certified) {
          done();
        }
      };
      getSnsNeuron({
        neuronIdHex: bytesToHexString(
          Array.from(mockSnsNeuron.id[0]?.id as Uint8Array)
        ),
        rootCanisterId: mockPrincipal,
        onLoad,
      });
    });

    it("should call api even if it's in the store when forceFetch", (done) => {
      const spyQuery = jest
        .spyOn(api, "getSnsNeuron")
        .mockImplementation(() => Promise.resolve({ ...mockSnsNeuron }));
      snsNeuronsStore.setNeurons({
        rootCanisterId: mockPrincipal,
        neurons: [mockSnsNeuron],
        certified: true,
      });
      const onLoad = ({
        neuron,
        certified,
      }: {
        neuron: SnsNeuron;
        certified: boolean;
      }) => {
        expect(spyQuery).toBeCalled();
        expect(neuron).not.toBe(mockSnsNeuron);
        if (certified) {
          done();
        }
      };
      getSnsNeuron({
        forceFetch: true,
        neuronIdHex: bytesToHexString(
          Array.from(mockSnsNeuron.id[0]?.id as Uint8Array)
        ),
        rootCanisterId: mockPrincipal,
        onLoad,
      });
    });

    it("should call onError callback when call failes", (done) => {
      const spyQuery = jest
        .spyOn(api, "getSnsNeuron")
        .mockImplementation(() => Promise.reject());
      const onLoad = jest.fn();
      const onError = ({ certified }: { certified: boolean }) => {
        expect(spyQuery).toBeCalled();
        expect(onLoad).not.toBeCalled();
        if (certified) {
          done();
        }
      };
      getSnsNeuron({
        neuronIdHex: bytesToHexString(
          Array.from(mockSnsNeuron.id[0]?.id as Uint8Array)
        ),
        rootCanisterId: mockPrincipal,
        onLoad,
        onError,
      });
    });
  });

  describe("addHotkey", () => {
    it("should call api.addNeuronPermissions", async () => {
      const spyAdd = jest
        .spyOn(governanceApi, "addNeuronPermissions")
        .mockImplementation(() => Promise.resolve());
      const hotkey = Principal.fromText("aaaaa-aa");
      const { success } = await addHotkey({
        neuronId: mockSnsNeuron.id[0] as SnsNeuronId,
        hotkey,
        rootCanisterId: mockPrincipal,
      });
      expect(success).toBeTruthy();
      expect(spyAdd).toBeCalledWith({
        neuronId: mockSnsNeuron.id[0] as SnsNeuronId,
        identity: mockIdentity,
        principal: hotkey,
        rootCanisterId: mockPrincipal,
        permissions: HOTKEY_PERMISSIONS,
      });
    });
  });

  describe("removeHotkey", () => {
    it("should call api.addNeuronPermissions", async () => {
      const spyAdd = jest
        .spyOn(governanceApi, "removeNeuronPermissions")
        .mockImplementation(() => Promise.resolve());
      const hotkey = "aaaaa-aa";
      const { success } = await removeHotkey({
        neuronId: mockSnsNeuron.id[0] as SnsNeuronId,
        hotkey,
        rootCanisterId: mockPrincipal,
      });
      expect(success).toBeTruthy();
      expect(spyAdd).toBeCalledWith({
        neuronId: mockSnsNeuron.id[0] as SnsNeuronId,
        identity: mockIdentity,
        principal: Principal.fromText(hotkey),
        rootCanisterId: mockPrincipal,
        permissions: HOTKEY_PERMISSIONS,
      });
    });
  });

  describe("disburse", () => {
    it("should call api.disburse", async () => {
      const neuronId = mockSnsNeuron.id[0] as SnsNeuronId;
      const identity = mockIdentity;
      const rootCanisterId = mockPrincipal;

      const spyOnDisburse = jest
        .spyOn(governanceApi, "disburse")
        .mockImplementation(() => Promise.resolve());

      const { success } = await disburse({
        rootCanisterId,
        neuronId,
      });

      expect(success).toBeTruthy();

      expect(spyOnDisburse).toBeCalledWith({
        neuronId,
        identity,
        rootCanisterId,
      });
    });
  });

  describe("start dissolving", () => {
    it("should call sns api startDissolving", async () => {
      const neuronId = mockSnsNeuron.id[0] as SnsNeuronId;
      const identity = mockIdentity;
      const rootCanisterId = mockPrincipal;

      const spyOnStartDissolving = jest
        .spyOn(governanceApi, "startDissolving")
        .mockImplementation(() => Promise.resolve());

      const { success } = await startDissolving({
        rootCanisterId,
        neuronId,
      });

      expect(success).toBeTruthy();

      expect(spyOnStartDissolving).toBeCalledWith({
        neuronId,
        identity,
        rootCanisterId,
      });
    });
  });

  describe("stop dissolving", () => {
    it("should call sns api stopDissolving", async () => {
      const neuronId = mockSnsNeuron.id[0] as SnsNeuronId;
      const identity = mockIdentity;
      const rootCanisterId = mockPrincipal;

      const spyOnStopDissolving = jest
        .spyOn(governanceApi, "stopDissolving")
        .mockImplementation(() => Promise.resolve());

      const { success } = await stopDissolving({
        rootCanisterId,
        neuronId,
      });

      expect(success).toBeTruthy();

      expect(spyOnStopDissolving).toBeCalledWith({
        neuronId,
        identity,
        rootCanisterId,
      });
    });
  });

  describe("updateDelay ", () => {
    const spyOnIncreaseDissolveDelay = jest
      .spyOn(governanceApi, "increaseDissolveDelay")
      .mockImplementation(() => Promise.resolve());

    beforeEach(spyOnIncreaseDissolveDelay.mockClear);

    it("should call sns api increaseDissolveDelay", async () => {
      const neuronId = fromDefinedNullable(mockSnsNeuron.id);
      const identity = mockIdentity;
      const rootCanisterId = mockPrincipal;
      const { success } = await updateDelay({
        rootCanisterId,
        dissolveDelaySeconds: 123,
        neuron: mockSnsNeuron,
      });

      expect(success).toBeTruthy();

      expect(spyOnIncreaseDissolveDelay).toBeCalledWith({
        neuronId,
        identity,
        rootCanisterId,
        additionalDissolveDelaySeconds: 123,
      });
    });

    it("should calculate additionalDissolveDelaySeconds", async () => {
      const rootCanisterId = mockPrincipal;
      const { success } = await updateDelay({
        rootCanisterId,
        dissolveDelaySeconds: 333,
        neuron: {
          ...mockSnsNeuron,
          dissolve_state: [{ DissolveDelaySeconds: BigInt(111) }],
        },
      });

      expect(success).toBeTruthy();

      expect(spyOnIncreaseDissolveDelay).toHaveBeenNthCalledWith(
        1,
        expect.objectContaining({
          additionalDissolveDelaySeconds: 222,
        })
      );
    });
  });

  describe("stakeNeuron", () => {
    afterEach(() => {
      transactionsFeesStore.reset();
      jest.clearAllMocks();
    });

    it("should call sns api stakeNeuron, query neurons again and load sns accounts", async () => {
      transactionsFeesStore.setFee({
        rootCanisterId: mockPrincipal,
        fee: BigInt(100),
        certified: true,
      });
      const spyStake = jest
        .spyOn(api, "stakeNeuron")
        .mockImplementation(() => Promise.resolve(mockSnsNeuron.id[0]));
      const spyQuery = jest
        .spyOn(api, "querySnsNeurons")
        .mockImplementation(() => Promise.resolve([mockSnsNeuron]));

      const { success } = await stakeNeuron({
        rootCanisterId: mockPrincipal,
        amount: 2,
        account: mockSnsMainAccount,
      });

      expect(success).toBeTruthy();
      expect(spyStake).toBeCalled();
      expect(spyQuery).toBeCalled();
      expect(loadSnsAccounts).toBeCalled();
    });

    it("should not call sns api stakeNeuron if fee is not present", async () => {
      transactionsFeesStore.reset();
      const spyStake = jest
        .spyOn(api, "stakeNeuron")
        .mockImplementation(() => Promise.resolve(mockSnsNeuron.id[0]));

      const { success } = await stakeNeuron({
        rootCanisterId: mockPrincipal,
        amount: 2,
        account: mockSnsMainAccount,
      });

      expect(success).toBeFalsy();
      expect(spyStake).not.toBeCalled();
    });
  });

  describe("increaseStakeNeuron", () => {
    it("should call api.increaseStakeNeuron and load sns accounts", async () => {
      const spyOnIncreaseStakeNeuron = jest
        .spyOn(api, "increaseStakeNeuron")
        .mockImplementation(() => Promise.resolve());

      const rootCanisterId = mockPrincipal;
      const amount = 2;
      const identity = mockIdentity;
      const neuronId = mockSnsNeuron.id[0] as SnsNeuronId;
      const account = mockSnsMainAccount;
      const identifier = decodeIcrcAccount(account.identifier);

      const { success } = await increaseStakeNeuron({
        rootCanisterId,
        amount,
        account,
        neuronId,
      });

      expect(success).toBeTruthy();

      expect(spyOnIncreaseStakeNeuron).toBeCalledWith({
        neuronId,
        rootCanisterId,
        stakeE8s: numberToE8s(amount),
        identity,
        source: identifier,
      });
      expect(loadSnsAccounts).toBeCalled();
    });
  });

  describe("stakeMaturity", () => {
    it("should call api.stakeMaturity", async () => {
      const neuronId = mockSnsNeuron.id[0] as SnsNeuronId;
      const identity = mockIdentity;
      const rootCanisterId = mockPrincipal;
      const percentageToStake = 60;

      const spyOnStakeMaturity = jest
        .spyOn(governanceApi, "stakeMaturity")
        .mockImplementation(() => Promise.resolve());

      const { success } = await stakeMaturity({
        neuronId,
        rootCanisterId,
        percentageToStake,
      });

      expect(success).toBeTruthy();

      expect(spyOnStakeMaturity).toBeCalledWith({
        neuronId,
        rootCanisterId,
        percentageToStake,
        identity,
      });
    });
  });

  describe("addFollowee ", () => {
    const setFolloweesSpy = jest
      .spyOn(governanceApi, "setFollowees")
      .mockImplementation(() => Promise.resolve());

    const followee1: SnsNeuronId = {
      id: arrayOfNumberToUint8Array([1, 2, 3]),
    };
    const followee2: SnsNeuronId = {
      id: arrayOfNumberToUint8Array([1, 2, 4]),
    };
    const followeeHex2 = subaccountToHexString(followee2.id);
    const rootCanisterId = mockPrincipal;
    const functionId = BigInt(3);

    afterEach(() => jest.clearAllMocks());

    it("should call sns api setFollowees with new followee when topic already has followees", async () => {
      const queryNeuronSpy = jest
        .spyOn(api, "querySnsNeuron")
        .mockResolvedValue(mockSnsNeuron);
      const neuron: SnsNeuron = {
        ...mockSnsNeuron,
        followees: [[functionId, { followees: [followee1] }]],
      };
      await addFollowee({
        rootCanisterId,
        neuron,
        functionId,
        followeeHex: followeeHex2,
      });

      expect(setFolloweesSpy).toBeCalledWith({
        neuronId: fromNullable(neuron.id),
        identity: mockIdentity,
        rootCanisterId,
        followees: [followee1, followee2],
        functionId,
      });
      expect(queryNeuronSpy).toBeCalled();
    });

    it("should call sns api setFollowees with new followee when topic has no followees", async () => {
      jest.spyOn(api, "querySnsNeuron").mockResolvedValue(mockSnsNeuron);
      const neuron: SnsNeuron = {
        ...mockSnsNeuron,
        followees: [[BigInt(4), { followees: [followee1] }]],
      };
      await addFollowee({
        rootCanisterId,
        neuron,
        functionId,
        followeeHex: followeeHex2,
      });

      expect(setFolloweesSpy).toBeCalledWith({
        neuronId: fromNullable(neuron.id),
        identity: mockIdentity,
        rootCanisterId,
        followees: [followee2],
        functionId,
      });
    });

    it("should not call sns api setFollowees when new followee is in the list", async () => {
      jest.spyOn(api, "querySnsNeuron").mockResolvedValue(mockSnsNeuron);
      const neuron: SnsNeuron = {
        ...mockSnsNeuron,
        followees: [[functionId, { followees: [followee2] }]],
      };
      await addFollowee({
        rootCanisterId,
        neuron,
        functionId,
        followeeHex: followeeHex2,
      });

      expect(setFolloweesSpy).not.toBeCalled();
      expect(toastsError).toBeCalled();
    });

    it("should not call sns api setFollowees when new followee is the same neuron", async () => {
      jest.spyOn(api, "querySnsNeuron").mockResolvedValue(mockSnsNeuron);
      const neuronIdHext = getSnsNeuronIdAsHexString(mockSnsNeuron);
      await addFollowee({
        rootCanisterId,
        neuron: mockSnsNeuron,
        functionId,
        followeeHex: neuronIdHext,
      });

      expect(setFolloweesSpy).not.toBeCalled();
      expect(toastsError).toBeCalled();
    });

    it("should not call sns api setFollowees when new followee does not exist", async () => {
      jest.spyOn(api, "querySnsNeuron").mockResolvedValue(undefined);
      const neuronIdHext = getSnsNeuronIdAsHexString(mockSnsNeuron);
      await addFollowee({
        rootCanisterId,
        neuron: mockSnsNeuron,
        functionId,
        followeeHex: neuronIdHext,
      });

      expect(setFolloweesSpy).not.toBeCalled();
      expect(toastsError).toBeCalled();
    });
  });

  describe("removeFollowee ", () => {
    const setFolloweesSpy = jest
      .spyOn(governanceApi, "setFollowees")
      .mockImplementation(() => Promise.resolve());

    const followee1: SnsNeuronId = {
      id: arrayOfNumberToUint8Array([1, 2, 3]),
    };
    const followee2: SnsNeuronId = {
      id: arrayOfNumberToUint8Array([1, 2, 4]),
    };
    const rootCanisterId = mockPrincipal;
    const functionId = BigInt(3);

    afterEach(() => jest.clearAllMocks());

    it("should call sns api setFollowees with followee removed from list", async () => {
      const neuron: SnsNeuron = {
        ...mockSnsNeuron,
        followees: [[functionId, { followees: [followee1, followee2] }]],
      };
      await services.removeFollowee({
        rootCanisterId,
        neuron,
        functionId,
        followee: followee1,
      });

      expect(setFolloweesSpy).toBeCalledWith({
        neuronId: fromNullable(neuron.id),
        identity: mockIdentity,
        rootCanisterId,
        followees: [followee2],
        functionId,
      });
    });

    it("should call sns api setFollowees with empty list if followee is the last followee", async () => {
      const neuron: SnsNeuron = {
        ...mockSnsNeuron,
        followees: [[functionId, { followees: [followee1] }]],
      };
      await services.removeFollowee({
        rootCanisterId,
        neuron,
        functionId,
        followee: followee1,
      });

      expect(setFolloweesSpy).toBeCalledWith({
        neuronId: fromNullable(neuron.id),
        identity: mockIdentity,
        rootCanisterId,
        followees: [],
        functionId,
      });
    });

    it("should not call sns api setFollowees when followee is not in the list", async () => {
      jest.spyOn(api, "querySnsNeuron").mockResolvedValue(mockSnsNeuron);
      const neuron: SnsNeuron = {
        ...mockSnsNeuron,
        followees: [[functionId, { followees: [followee2] }]],
      };
      await services.removeFollowee({
        rootCanisterId,
        neuron,
        functionId,
        followee: followee1,
      });

      expect(setFolloweesSpy).not.toBeCalled();
      expect(toastsError).toBeCalled();
    });
  });

  describe("stakeMaturity", () => {
    it("should call api.stakeMaturity", async () => {
      const neuronId = mockSnsNeuron.id[0] as SnsNeuronId;
      const identity = mockIdentity;
      const rootCanisterId = mockPrincipal;
      const percentageToStake = 60;

      const spyOnStakeMaturity = jest
        .spyOn(governanceApi, "stakeMaturity")
        .mockImplementation(() => Promise.resolve());

      const { success } = await stakeMaturity({
        neuronId,
        rootCanisterId,
        percentageToStake,
      });

      expect(success).toBeTruthy();

      expect(spyOnStakeMaturity).toBeCalledWith({
        neuronId,
        rootCanisterId,
        percentageToStake,
        identity,
      });
    });
  });

  describe("toggleAutoStakeMaturity", () => {
    const spyOnStakeMaturity = jest
      .spyOn(governanceApi, "autoStakeMaturity")
      .mockImplementation(() => Promise.resolve());

    const testToggle = async ({
      neuron,
      neuronId,
      rootCanisterId,
      identity,
      autoStake,
    }) => {
      const { success } = await toggleAutoStakeMaturity({
        neuron,
        neuronId,
        rootCanisterId,
      });

      expect(success).toBeTruthy();

      expect(spyOnStakeMaturity).toBeCalledWith({
        neuronId,
        rootCanisterId,
        identity,
        autoStake,
      });
    };

    const neuronId = mockSnsNeuron.id[0] as SnsNeuronId;
    const identity = mockIdentity;
    const rootCanisterId = mockPrincipal;

    it("should call api.autoStakeMaturity with true for the first toggle", async () => {
      const neuron = {
        ...mockSnsNeuron,
        auto_stake_maturity: [] as [] | [boolean],
      };

      await testToggle({
        neuronId,
        neuron,
        identity,
        rootCanisterId,
        autoStake: true,
      });
    });

    it("should call api.autoStakeMaturity with false", async () => {
      const neuron = {
        ...mockSnsNeuron,
        auto_stake_maturity: [true] as [] | [boolean],
      };

      await testToggle({
        neuronId,
        neuron,
        identity,
        rootCanisterId,
        autoStake: false,
      });
    });

    it("should call api.autoStakeMaturity with true", async () => {
      const neuron = {
        ...mockSnsNeuron,
        auto_stake_maturity: [false] as [] | [boolean],
      };

      await testToggle({
        neuronId,
        neuron,
        identity,
        rootCanisterId,
        autoStake: true,
      });
    });
  });

  describe("splitNeuron", () => {
    const transactionFee = 100n;
    let snsNeuronsStoreSpy: jest.SpyInstance;
    let snsTokenSymbolSelectedStoreSpy: jest.SpyInstance;

    beforeEach(() => {
      snsNeuronsStoreSpy = jest
        .spyOn(snsNeuronsStore, "subscribe")
        .mockImplementation(
          buildMockSnsNeuronsStoreSubscribe({
            rootCanisterId: mockPrincipal,
            neurons: [mockSnsNeuron],
          })
        );
      snsTokenSymbolSelectedStoreSpy = jest
        .spyOn(snsTokenSymbolSelectedStore, "subscribe")
        .mockImplementation(mockTokenStore);

      transactionsFeesStore.setFee({
        rootCanisterId: mockPrincipal,
        fee: BigInt(transactionFee),
        certified: true,
      });
    });

    afterEach(() => {
      snsNeuronsStoreSpy.mockClear();
      snsTokenSymbolSelectedStoreSpy.mockClear();
      transactionsFeesStore.reset();
    });

    it("should call api.addNeuronPermissions", async () => {
      const spySplitNeuron = jest
        .spyOn(governanceApi, "splitNeuron")
        .mockImplementation(() => Promise.resolve());
      const spyLoadNeurons = jest
        .spyOn(services, "loadNeurons")
        .mockResolvedValue(undefined);
      const amount = 10;

      const neuronMinimumStake = 1000n;
      const { success } = await splitNeuron({
        neuronId: mockSnsNeuron.id[0] as SnsNeuronId,
        rootCanisterId: mockPrincipal,
        amount,
        neuronMinimumStake,
      });
      expect(success).toBeTruthy();
      expect(spyLoadNeurons).toBeCalled();
      expect(spySplitNeuron).toBeCalledWith({
        neuronId: mockSnsNeuron.id[0] as SnsNeuronId,
        identity: mockIdentity,
        rootCanisterId: mockPrincipal,
        amount: BigInt(amount * E8S_PER_ICP) + transactionFee,
        memo: 0n,
      });
    });

    it("should display error if not enough amount", async () => {
      const spySplitNeuron = jest
        .spyOn(governanceApi, "splitNeuron")
        .mockImplementation(() => Promise.resolve())
        .mockReset();
      const amount = 0.00001;
      const neuronMinimumStake = 2000n;
      const { success } = await splitNeuron({
        neuronId: mockSnsNeuron.id[0] as SnsNeuronId,
        rootCanisterId: mockPrincipal,
        amount,
        neuronMinimumStake,
      });

      expect(toastsError).toBeCalled();
      expect(success).toBeFalsy();
      expect(spySplitNeuron).not.toBeCalled();
    });
  });
});
